from __future__ import annotations
from functools import wraps
from inspect import signature
from .labeledarray import LabeledArray
from .label import Label
from .specials import PropArray
from .deco import dims_to_spatial_axes, make_history
import numpy as np
import os
import glob
import itertools
import collections
from skimage import io
from .imgarray import ImgArray
from .func import *
from .bases import MetaArray, HistoryArray
from .axes import Axes, ImageAxesError
from .utilcls import Progress
from skimage import data as skdata
from typing import Callable

def array(arr, dtype=None, *, name=None, axes=None) -> ImgArray:
    """
    make an ImgArray object, just like np.array(x)
    """
    if isinstance(arr, str):
        raise TypeError(f"String is invalid input. Do you mean imread(path)?")
    if isinstance(arr, np.ndarray) and dtype is None:
        if arr.dtype in (np.uint8, np.uint16, np.float32):
            dtype = arr.dtype
        elif arr.dtype.kind == "f":
            dtype = np.float32
        else:
            dtype = arr.dtype
    
    arr = np.asarray(arr, dtype=dtype)
        
    # Automatically determine axes
    if axes is None:
        axes = ["x", "yx", "tyx", "tzyx", "tzcyx", "ptzcyx"][arr.ndim-1]
            
    self = ImgArray(arr, name=name, axes=axes)
    
    return self

def zeros(shape, dtype=np.uint16, *, name=None, axes=None) -> ImgArray:
    return array(np.zeros(shape, dtype=dtype), dtype=dtype, name=name, axes=axes)

def empty(shape, dtype=np.uint16, *, name=None, axes=None) -> ImgArray:
    return array(np.empty(shape, dtype=dtype), dtype=dtype, name=name, axes=axes)

def imread(path:str, dtype:str=None, *, axes=None) -> ImgArray:
    """
    Load image from path.

    Parameters
    ----------
    path : str
        Path to the image.
    dtype : Any type that np.dtype accepts
        dtype of the images.
    axes : str or None, optional
        If the image does not have axes metadata, this value will be used.

    Returns
    -------
    ImgArray
    """    
    if not os.path.exists(path):
        raise FileNotFoundError(f"No such file or directory: {path}")
    
    fname, fext = os.path.splitext(os.path.basename(path))
    img = io.imread(path)
    dirpath = os.path.dirname(path)
    
    # read tif metadata
    if fext == ".tif":
        meta = get_meta(path)
    elif fext in (".png", ".jpg") and img.ndim == 3 and img.shape[-1] <= 4:
        meta = {"axes":"yxc", "ijmeta":{}, "history":[]}
    else:
        meta = {"axes":axes, "ijmeta":{}, "history":[]}
    
    axes = meta["axes"]
    metadata = meta["ijmeta"]
    if meta["history"]:
        name = meta["history"].pop(0)
        history = meta["history"]
    else:
        name = fname
        history = []
    
    self = ImgArray(img, name=name, axes=axes, dirpath=dirpath, 
                    history=history, metadata=metadata)
        
    # In case the image is in yxc-order. This sometimes happens.
    if "c" in self.axes and self.sizeof("c") > self.sizeof("x"):
        self = np.moveaxis(self, -1, -3)
        _axes = self.axes.axes
        _axes = _axes[:-3] + "cyx"
        self.axes = _axes
    
    if dtype is None:
        dtype = self.dtype
        
    if self.axes.is_none():
        return self
    else:
        # read lateral scale if possible
        spacing = meta["ijmeta"].get("spacing", 1.0)
        try:
            tags = meta["tags"]
            xres = tags["XResolution"]
            yres = tags["YResolution"]
            dx = xres[1]/xres[0]
            dy = yres[1]/yres[0]
        except KeyError:
            dx = dy = spacing
        
        self.set_scale(x=dx, y=dy)
        
        # read z scale if needed
        if "z" in self.axes:
            dz = spacing
            self.set_scale(z=dz)
        return self.sort_axes().as_img_type(dtype) # arrange in tzcyx-order

def imread_collection(dirname:str, axis:str="p", *, filename:str="*.tif", template:dict|MetaArray=None,
                      ignore_exception:bool=False, dtype=None) -> ImgArray:
    """
    Read images recursively from a directory, and stack them into one ImgArray.

    Parameters
    ----------
    dirname : str
        Path to the directory
    axis : str, default is "p"
        To specify which axis will be the new one.
    filname : str, default is "*.tif"
        File name that satisfies this string will be read. This variable will be passed to `glob.glob`.
    template : dict or MetaArray, optional
        Images that matches the template will added to image stack.
    ignore_exception : bool, default is False
        If true, arrays with wrong shape will be ignored.
    dtype : Any type that np.dtype accepts
        dtype of the images.
    
    Example
    -------
    (1) Read Tiff images that start with "100nM-":
    >>> img = ip.imread_collection(r"C:\...", filename="100nM-*.tif")
    
    (2) Read Tiff images that have tyx-axes:
    >>> img = ip.imread_collection(r"C:\...", template={"axes: "tyx"})
    
    (3) Read Tiff images that have strictly same features as a reference image `ref`:
    >>> img = ip.imread_collection(r"C:\...", template=ref)
    """    
    paths = glob.glob(os.path.join(dirname, "**", filename), recursive=True)
    
    # determine template
    template_keys = {"shape", "axes", "scale"}
    if template is None:
        template = {}
    elif isinstance(template, dict):
        if not set(template.keys()) <= template_keys:
            raise ValueError(f"template only takes {template_keys} as keys.")
    elif isinstance(template, MetaArray):
        template = {k: getattr(template, k) for k in template_keys}
    else:
        raise TypeError(f"template must be dict or MetaArray, but got {type(template)}.")
    
    imgs = []
    shapes = []
    for path in paths:
        img = imread(path, dtype=dtype)
        for k, v in template.items():
            if getattr(img, k) != v:
                continue
        imgs.append(img)
        shapes.append(img.shape)
    
    # check shape compatibility
    list_of_shape = list(set(shapes))
    if len(list_of_shape) > 1:
        if ignore_exception:
            ctr = collections.Counter(shapes)
            common_shape = ctr.most_common()[0][0]
            imgs = [img for img in imgs if img.shape == common_shape]
        else:
            raise ValueError("Input directory has images with different shapes: "
                            f"{', '.join(map(str, list_of_shape))}")
    
    if len(imgs) == 0:
        raise RuntimeError("Could not read any images.")
    
    out = stack(imgs, axis=axis)
    out.dirpath, out.name = os.path.split(dirname)
    out.history[-1] = "imread_collection"
    out.temp = paths
    return out

def imread_stack(path:str, dtype=None):
    r"""
    Read separate image files using formated string. This function is useful when files/folders
    are named in a certain rule, such as ".../pos_0/img_0.tif", ".../pos_0/img_1.tif".

    Parameters
    ----------
    path : str
        Formated path string.
    dtype : Any type that np.dtype accepts
        dtype of the images.
    
    Returns
    -------
    ImgArray
        Image stack
    
    Example
    -------
    (1) For following file structure, read pos0, pos1, ... as p-stack.
        Base
        |- pos0.tif
        |- pos1.tif
        |- pos2.tif
        :
    >>> img = ip.imread_stack(r"C:\...\Base\pos$p.tif")
    
    (2) For following file structure, read xxx0, xxx1, ... as z-stack, and read yyy0, yyy1, ...
    as t-stack.
        Base
        |- xxx0
        |   |- yyy0.tif
        |   |- yyy1.tif
        |       :
        |- xxx1
        |   |- yyy0.tif
        |   |- yyy1.tif
        |       :
        :
    >>> img = ip.imread_stack(r"C:\...\Base\xxx$z\yyy$t.tif")
    """
    FORMAT = r"\$[a-z]"
    new_axes = list(map(lambda x: x[1:], re.findall(r"\$[a-z]", path)))
    
    # To convert input path string into that for glob.glob, replace $X with wildcard
    # e.g.) ~\XX$t_YY$z -> ~\XX*_YY*
    finder_path = re.sub(FORMAT, "*", path)
    
    # To convert input path string into file-finding regex pattern.
    # e.g.) ~\XX$t_YY$z\*.tif -> ~\\XX(\d)_YY(\d)\\.*\.tif
    path = repr(path)[1:-1]
    pattern = re.sub(r"\.", r"\.",path)          # dots to non-escape
    pattern = re.sub(r"\*", ".*", pattern)       # asters to non-escape
    pattern = re.sub(FORMAT, r"(\\d+)", pattern) # make number finders
    pattern = re.compile(pattern)
    
    # To convert input path string into format string to execute imread.
    # e.g.) ~\XX$t_YY$z -> ~\XX{}_YY{}
    fpath = re.sub(FORMAT, "{}", path)
    
    paths = glob.glob(finder_path)
    indices = [pattern.findall(p) for p in paths]
    ranges = [list(np.unique(ind)) for ind in np.array(indices).T]
    ranges_sorted = [[r[i] for i in np.argsort(list(map(int, r)))] for r in ranges]
    
    # read all the images
    img0 = None
    imgs = []
    for i in itertools.product(*ranges_sorted):
        # check if the image to read is unique
        found_paths = glob.glob(fpath.format(*i))
        n_found = len(found_paths)
        if n_found > 1:
            raise ValueError(f"{n_found} paths found at {fpath.format(*i)}.")
        elif n_found == 0:
            raise FileNotFoundError(f"No path found at {fpath.format(*i)}.")
        
        img = imread(found_paths[0], dtype=dtype)
        
        # To speed up error handling, check shape and axes here.
        if img0 is None:
            img0 = img
            for a in new_axes:
                if a in img0.axes:
                    raise ImageAxesError(f"{a} appeared twice.")
        else:
            if img.shape != img0.shape:
                raise ValueError(f"Shape mismatch at {fpath.format(*i)}. Make sure all "
                                  "the input images have exactly the same shapes.")
                
        imgs.append(img)
    
    # reshape image and set metadata
    new_shape = tuple(len(r) for r in ranges) + imgs[0].shape
    self = np.array(imgs, dtype=dtype).reshape(*new_shape).view(ImgArray)
    self._set_info(imgs[0])
    self.axes = "".join(new_axes) + str(img.axes)
    self.set_scale(imgs[0])
    return self.sort_axes()
    

def read_meta(path:str) -> dict[str]:
    """
    Read the metadata of a tiff file. 

    Parameters
    ----------
    path : str
        Path to the tiff file.

    Returns
    -------
    dict
        Dictionary of metadata with following keys.
        "axes": axes information
        "ijmeta": ImageJ metadata
        "history": impy history
        "tags": tiff tags
    """    
    if not path.endswith(".tif"):
        raise ValueError("Cannot read metadata from file extension other than tif.")
    meta = get_meta(path)
    return meta

def stack(imgs, axis="c", dtype=None):
    raise Exception("use np.stack instead.")

def set_cpu(n_cpu:int) -> None:
    ImgArray.n_cpu = n_cpu
    return None

def set_verbose(b:bool) -> None:
    Progress.show_progress = b
    return None

def sample_image(name:str) -> ImgArray:
    """
    Get sample images from `skimage` and convert it into ImgArray.

    Parameters
    ----------
    name : str
        Name of sample image, such as "camera".

    Returns
    -------
    ImgArray
        Sample image.
    """    
    img = getattr(skdata, name)()
    out = array(img, name=name)
    if out.shape[-1] == 3:
        out.axes = "yxc"
        out = out.sort_axes()
    return out

# Overloading numpy functions using __array_function__.
# https://numpy.org/devdocs/reference/arrays.classes.html

def safe_set_info(out, img, history, new_axes):
    if isinstance(img, HistoryArray):
        out._set_info(img, history, new_axes=new_axes)
    else:
        try:
            out._set_info(img, new_axes=new_axes)
        except Exception:
            pass
    return None

@MetaArray.implements(np.squeeze)
def squeeze(img:MetaArray):
    out = np.squeeze(img.value).view(img.__class__)
    new_axes = "".join(a for a in img.axes if img.sizeof(a) > 1)
    safe_set_info(out, img, "squeeze", new_axes)
    return out

@MetaArray.implements(np.take)
def take(a:MetaArray, indices, axis:int=None, out=None, mode="raise"):
    new_axes = del_axis(a.axes, axis)
    if isinstance(axis, str):
        axis = a.axes.find(axis)
    out = np.take(a.value, indices, axis=axis, out=out, mode=mode).view(a.__class__)
    if isinstance(out, a.__class__):
        out._set_info(a, new_axes=new_axes)
    return out

@MetaArray.implements(np.stack)
def stack(imgs:list[MetaArray], axis="c", dtype=None):
    """
    Create stack image from list of images.

    Parameters
    ----------
    imgs : iterable object of images.
        Images to stack. These images must have exactly the same shapes.
    axis : str, default is "c"
        Which axis will be the new one.
    dtype : str, optional
        Output dtype.

    Returns
    -------
    ImgArray
        Image stack
    """
    if not isinstance(axis, (int, str)):
        raise TypeError(f"`axis` must be int or str, but got {type(axis)}")
    
    # find where to add new axis
    if imgs[0].axes:
        new_axes = Axes(axis + str(imgs[0].axes))
        new_axes.sort()
        _axis = new_axes.find(axis)
    else:
        new_axes = None
        _axis = 0

    if dtype is None:
        dtype = imgs[0].dtype

    arrs = [img.as_img_type(dtype).value for img in imgs]

    out = np.stack(arrs, axis=0)
    out = np.moveaxis(out, 0, _axis)
    out = out.view(imgs[0].__class__)
    safe_set_info(out, imgs[0], f"stack(axis={axis})", new_axes)
    return out

@MetaArray.implements(np.concatenate)
def concatenate(imgs, axis="c", dtype=None, casting="same_kind"):
    if not isinstance(axis, (int, str)):
        raise TypeError(f"`axis` must be int or str, but got {type(axis)}")
    axis = imgs[0].axisof(axis)
    out = np.concatenate([img.value for img in imgs], axis=axis, dtype=dtype, casting=casting)
    out = out.view(imgs[0].__class__)
    safe_set_info(out, imgs[0], f"concatenate(axis={axis})", imgs[0].axes)
    return out

@MetaArray.implements(np.block)
def block(imgs):
    def _recursive_view(obj):
        if isinstance(obj, MetaArray):
            return obj.value
        else:
            return [_recursive_view(a) for a in obj]
    
    def _recursive_get0(obj):
        first = obj[0]
        if isinstance(first, MetaArray):
            return first
        else:
            return _recursive_get0(first)
    
    img0 = _recursive_get0(imgs)
    
    imgs = _recursive_view(imgs)
    out = np.block(imgs).view(img0.__class__)
    safe_set_info(out, img0, "block", img0.axes)
    return out


@MetaArray.implements(np.zeros_like)
def zeros_like(img, name:str=None):
    out = np.zeros_like(img.value).view(img.__class__)
    out._set_info(img, new_axes=img.axes)
    if isinstance(name, str):
        out.name = name
    return out

@MetaArray.implements(np.empty_like)
def empty_like(img, name:str=None):
    out = np.empty_like(img.value).view(img.__class__)
    out._set_info(img, new_axes=img.axes)
    if isinstance(name, str):
        out.name = name
    return out

@MetaArray.implements(np.expand_dims)
def expand_dims(img, axis):
    if isinstance(axis, str):
        new_axes = Axes(axis + str(img.axes))
        new_axes.sort()
        axisint = tuple(new_axes.find(a) for a in axis)
    else:
        axisint = axis
        new_axes = img.axes
    
    out = np.expand_dims(img.value, axisint).view(img.__class__)
    safe_set_info(out, img, f"expand_dims({axis})", new_axes)
    return out

@MetaArray.implements(np.transpose)
def transpose(img, axes):
    return img.transpose(axes)


# Extend ImgArray with costum functions.

class bind:
    """
    Dynamically define ImgArray function that can iterate over axes. You can integrate your own
    function, or useful functions from `skimage` or `opencv`. History of function call will be 
    similarly recorded in `self.history`.
    This class is designed as a kind of decorator class so that it can be used as decorator of
    any function or directly takes a function as the first argument.

    Parameters
    ----------
    func : callable
        Function to wrapped and bound to ImgArray.
    funcname : str, optional
        Method name after set to ImgArray. The name of `func` itself will be set by default.
    indtype : dtype, optional
        If given, input data type will be converted by `as_img_type` method before passed to `func`.
    outdtype : dtype, optional
        If given, output data array will be defined in this type if needed.
    kind : str, {"image", "property", "label", "label_binary"}, default is "image"
        What kind of function will be bound.
        - "image" ... Given an image, calculate a new image that has the exactily same shape.
          Bound method will return `ImgArray` that has the same shape and axes as the input image.
        - "property" ... Given an image, calculate a scalar value or any other object such as
        tuple, and store them in a `PropArray`. Axes of returned `PropArray` is (the axes of input
        image) - (axes of spatial dimensions specified by `dims` argument of bound method).
        - "label" ... Given an image, calculate a label image with value 0 being background and set
        it to `labels` attribute. The label image must have the exactly same shape as input image.
        - "label_binary" ... Given an image, calculate a binary image. Label image is generated from
        the binary image with `label` method in `LabeledArray`. The connectivity is None. The binary
        image must have the exactly same shape as input image.
    ndim : {None, 2, 3}, default is None
        Dimension of image that the original function supports. If None, then it is assumed to
        support both 2 and 3 dimensional images and automatically determined by the universal
        `dims_to_spatial_axes` method.
    mapping : dict of str -> (str, callable), optional
        If given, keyword arguments are converted using this mapping before passed to `func`. This
        keyword is used for modifing original function without wrapping it. For more detail see
        Example (2).

    Examples
    --------
    (1) Bind "normalize" method that will normalize images separately.
    >>> def normalize(img):
    >>>    min_, max_ = img.min(), img.max()
    >>>    return (img - min_)/(max_ - min_)
    >>> ip.bind(normalize, indtype=np.float32, outdtype=np.float32)
    >>> img = ip.imread(...)
    >>> img.normalize()
    
    (2) Bind `skimage.filters.rank.maximum` for filtering, but make it take "radius" rather than
    "selem" as a keyword argument.
    >>> from impy.func import ball_like
    >>> from skimage.filters.rank import maximum
    >>> ip.bind(maximum, "max_filter", mapping={"radius":("selem", ball_like)})
    >>> img = ip.imread(...)
    >>> img.max_filter(radius=3)
    
    (3) Bind a method `calc_mean` that calculate mean value around spatial dimensions. For one yx-
    or zyx-image, a scalar value is returned, so that `calc_mean` should return `PropArray`.
    >>> ip.bind(np.mean, "calc_mean", outdtype=np.float32, kind="property")
    >>> img = ip.imread(...)
    >>> img.calc_mean()
    
    (4) Wrap the normalize function in (1) in a decorator method.
    >>> @ip.bind(indtype=np.float32, outdtype=np.float32)
    >>> def normalize(img):
    >>>    min_, max_ = img.min(), img.max()
    >>>    return (img - min_)/(max_ - min_)
    >>> img = ip.imread(...)
    >>> img.normalize()
    or if you thick `indtype` and `outdtype` are unnecessary:
    >>> @ip.bind
    >>> def normalize(img):
    >>>     ...
    
    (5) Bind custom percentile labeling function (although `label_threshold` method can do the 
    exactly same thing).
    >>> @ip.bind(kind="label_binary")
    >>> def mylabel(img, p=90):
    >>>     per = np.percentile(img, p)
    >>>     thr = img > per
    >>>     return thr
    >>> img = ip.imread(...)
    >>> img.mylabel(95)   # img.labels is added here
    """    
    bound = set()
    def __init__(self, func:Callable=None, funcname:str=None, *, indtype=None, outdtype=None, 
                 kind:str="image", ndim:int|None=None, mapping:dict[str, tuple[str, Callable]]=None):
        """
        Method binding is done inside this when bind object is used as function like:
        >>> ip.bind(func, "funcname", ...)
        """        
        if callable(func):
            self._bind_method(func, funcname=funcname, indtype=indtype, outdtype=outdtype, 
                              kind=kind, ndim=ndim, mapping=mapping)
        else:
            self.funcname = func
            self.indtype = indtype
            self.outdtype = outdtype
            self.kind = kind
            self.ndim = ndim
            self.mapping = mapping
    
    def __call__(self, func:Callable):
        """
        Method binding is done inside this when bind object is used as decorator like:
        >>> @ip.bind(...)
        >>> def ...
        """
        if callable(func):
            self._bind_method(func, funcname=self.funcname, indtype=self.indtype, 
                              outdtype=self.outdtype, kind=self.kind, ndim=self.ndim, mapping=self.mapping)
        return func
    
    def _bind_method(self, func:Callable, funcname:str=None, *, indtype=None, outdtype=None,
                     kind="image", ndim=None, mapping:dict[str, tuple[str, Callable]]=None):
        # check function's name
        if funcname is None:
            fn = func.__name__
        elif isinstance(funcname, str):
            fn = funcname
        else:
            raise TypeError("`funcname` must be str if given.")
        if hasattr(ImgArray, fn) and fn not in self.bound:
            raise AttributeError(f"ImgArray already has attribute '{fn}'. Consider other names.")
        
        # check ndim and define default value of dims
        if ndim is None:
            default_dims = None
        elif ndim == 2:
            default_dims = "yx"
        elif ndim == 3:
            default_dims = "zyx"
        else:
            raise ValueError(f"`ndim` must be None, 2 or 3, but got {ndim}.")
        
        # check mapping
        if mapping is None:
            mapping = {}
        elif not isinstance(mapping, dict):
            raise TypeError(f"`mapping` must be dict, but got {type(mapping)}")
        
        # Dynamically define functions used inside the plugin method, depending on `kind` option.
        # _prepare_output_array : returns a subclass of ndarray for output.
        # _iter : returns an iterator around spatial dimensions.
        # _exit : overwrites output attributes.
        
        if kind == "image":
            def _prepare_output_array(self, dims):
                dtype = outdtype if outdtype is not None else self.dtype
                return np.empty(self.shape, dtype=dtype)
            
            def _iter(self, dims):
                return self.iter(complement_axes(dims, self.axes))
            
            def _exit(out, self, func, *args, **kwargs):
                out = out.view(ImgArray)
                history = make_history(func.__name__, args, kwargs)
                out._set_info(self, history)
                return out
            
        elif kind == "property":
            def _prepare_output_array(self, dims):
                dtype = outdtype if outdtype is not None else object
                c_axes = complement_axes(dims, self.axes)
                shape = self.sizesof(c_axes)
                return PropArray(np.empty(shape, dtype=dtype), name=self.name, dirpath=self.dirpath, 
                                axes=c_axes, dtype=dtype)
            
            def _iter(self, dims):
                return self.iter(complement_axes(dims, self.axes), exclude=dims)
            
            def _exit(out, self, func, *args, **kwargs):
                out.propname = fn
                return out
                
        elif kind == "label":
            def _prepare_output_array(self, dims):
                return largest_zeros(self.shape)
            
            def _iter(self, dims):
                return self.iter(complement_axes(dims, self.axes))
            
            def _exit(out, self, func, *args, **kwargs):
                self.labels = Label(out, name=self.name, axes=self.axes, dirpath=self.dirpath).optimize()
                self.labels.history.append(fn)
                self.labels.set_scale(self)
                return self.labels
            
        elif kind == "label_binary":
            def _prepare_output_array(self, dims):
                return largest_zeros(self.shape)
            
            def _iter(self, dims):
                return self.iter(complement_axes(dims, self.axes))
            
            def _exit(out, self, func, *args, **kwargs):
                self.labels = LabeledArray(out, axes=self.axes).label().labels
                self.labels.history.append(fn)
                self.labels.set_scale(self)
                return self.labels
                        
        else:
            raise NotImplementedError(kind)
        
        # Define method and bind it to ImgArray
        @wraps(func)
        @dims_to_spatial_axes
        def _func(self, *args, dims=default_dims, **kwargs):
            if indtype is not None:
                self = self.as_img_type(indtype)
            
            # map keyword arguments if necessary
            if mapping:
                kw = dict()
                for k, v in kwargs.items():
                    m = mapping.get(k, None)
                    if m is None:
                        kw[k] = v
                    else:
                        newkey, val = m
                        try:
                            kw[newkey] = val(v, len(dims))
                        except TypeError:
                            kw[newkey] = val(v)
                kwargs = kw
                
            out = _prepare_output_array(self, dims)
                
            with Progress(fn):
                for sl, img in _iter(self, dims):
                    out[sl] = func(img, *args, **kwargs)
                out = _exit(out, self, func, *args, **kwargs)
            return out
        
        self.__class__.bound.add(fn)
        return setattr(ImgArray, fn, _func)